cimport cpython.list
cimport cython
from libc.stdint cimport uint32_t


@cython.final
cdef class IProtoErrorStackFrame:
    def __repr__(self):
        return "<Frame type={} code={} message={}>".format(
            self.error_type,
            self.code,
            self.message,
        )

@cython.final
cdef class IProtoError:
    pass

cdef inline IProtoErrorStackFrame parse_iproto_error_stack_frame(const char ** b, bytes encoding):
    cdef:
        uint32_t size
        uint32_t key
        const char * s
        uint32_t s_len
        IProtoErrorStackFrame frame
        uint32_t unum

    size = 0
    key = 0

    frame = <IProtoErrorStackFrame> IProtoErrorStackFrame.__new__(IProtoErrorStackFrame)

    size = mp_decode_map(b)
    for _ in range(size):
        key = mp_decode_uint(b)

        if key == tarantool.MP_ERROR_TYPE:
            s = NULL
            s_len = 0
            s = mp_decode_str(b, &s_len)
            frame.error_type = decode_string(s[:s_len], encoding)

        elif key == tarantool.MP_ERROR_FILE:
            s = NULL
            s_len = 0
            s = mp_decode_str(b, &s_len)
            frame.file = decode_string(s[:s_len], encoding)

        elif key == tarantool.MP_ERROR_LINE:
            frame.line = <int> mp_decode_uint(b)

        elif key == tarantool.MP_ERROR_MESSAGE:
            s = NULL
            s_len = 0
            s = mp_decode_str(b, &s_len)
            frame.message = decode_string(s[:s_len], encoding)

        elif key == tarantool.MP_ERROR_ERRNO:
            frame.err_no = <int> mp_decode_uint(b)

        elif key == tarantool.MP_ERROR_ERRCODE:
            frame.code = <int> mp_decode_uint(b)

        elif key == tarantool.MP_ERROR_FIELDS:
            if mp_typeof(b[0][0]) != MP_MAP:  # pragma: nocover
                raise TypeError(f'iproto_error stack frame fields must be a '
                                f'map, but got {mp_typeof(b[0][0])}')

            frame.fields = _decode_obj(b, encoding)

        else:  # pragma: nocover
            logger.debug(f"unknown iproto_error stack element with key {key}")
            mp_next(b)

    return frame

cdef inline IProtoError iproto_error_decode(const char ** b, bytes encoding):
    cdef:
        uint32_t size
        uint32_t arr_size
        uint32_t key
        uint32_t i
        IProtoError error

    size = 0
    arr_size = 0
    key = 0

    error = <IProtoError> IProtoError.__new__(IProtoError)

    size = mp_decode_map(b)
    for _ in range(size):
        key = mp_decode_uint(b)

        if key == tarantool.MP_ERROR_STACK:
            arr_size = mp_decode_array(b)
            error.trace = cpython.list.PyList_New(arr_size)
            for i in range(arr_size):
                el = parse_iproto_error_stack_frame(b, encoding)
                cpython.Py_INCREF(el)
                cpython.list.PyList_SET_ITEM(error.trace, i, el)
        else:  # pragma: nocover
            logger.debug(f"unknown iproto_error map field with key {key}")
            mp_next(b)

    return error



from libc.stdint cimport uint32_t

from uuid import UUID


cdef object uuid_decode(const char ** p, uint32_t length):
    data = cpython.bytes.PyBytes_FromStringAndSize(p[0], length)
    p[0] += length
    return UUID(bytes=data)



from cpython.datetime cimport datetime
from libc.stdint cimport int16_t, int32_t, int64_t, uint32_t


cdef struct IProtoDateTime:
    int64_t seconds
    int32_t nsec
    int16_t tzoffset
    int16_t tzindex

cdef void datetime_zero(IProtoDateTime *dt)
cdef uint32_t datetime_len(IProtoDateTime *dt)
cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL
cdef int datetime_decode(const char ** p,
                         uint32_t length,
                         IProtoDateTime *dt) except -1
cdef void datetime_from_py(datetime ob, IProtoDateTime *dt)
cdef object datetime_to_py(IProtoDateTime *dt)



from libc cimport math
from libc.stdint cimport uint8_t, uint32_t


cdef inline uint32_t bcd_len(uint32_t digits_len):
    return <uint32_t> math.floor(digits_len / 2) + 1

cdef uint32_t decimal_len(int exponent, uint32_t digits_count)
cdef char *decimal_encode(char *p,
                          uint32_t digits_count,
                          uint8_t sign,
                          tuple digits,
                          int exponent) except NULL
cdef object decimal_decode(const char ** p, uint32_t length)



from libc.stdint cimport uint32_t

from decimal import Decimal


cdef uint32_t decimal_len(int exponent, uint32_t digits_count):
    cdef:
        uint32_t length

    length = bcd_len(digits_count)
    if exponent > 0:
        length += mp_sizeof_int(-exponent)
    else:
        length += mp_sizeof_uint(-exponent)

    return length

cdef char *decimal_encode(char *p,
                          uint32_t digits_count,
                          uint8_t sign,
                          tuple digits,
                          int exponent) except NULL:
    cdef:
        int i
        uint8_t byte
        char *out
        uint32_t length

    # encode exponent
    if exponent > 0:
        p = mp_encode_int(p, -exponent)
    else:
        p = mp_encode_uint(p, -exponent)

    length = bcd_len(digits_count)

    out = &p[length - 1]
    if sign == 1:
        byte = 0x0d
    else:
        byte = 0x0c

    i = digits_count - 1
    while out >= p:
        if i >= 0:
            byte |= (<uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i)) << 4

        out[0] = byte
        byte = 0

        if i > 0:
            byte = <uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i - 1) & 0xf

        out -= 1
        i -= 2

    p = &p[length]
    return p

cdef object decimal_decode(const char ** p, uint32_t length):
    cdef:
        int exponent
        uint8_t sign
        mp_type obj_type
        const char *first
        const char *last
        uint32_t digits_count
        uint8_t nibble

    sign = 0
    first = &p[0][0]
    last = first + length - 1

    # decode exponent
    obj_type = mp_typeof(p[0][0])
    if obj_type == MP_UINT:
        exponent = -<int> mp_decode_uint(p)
    elif obj_type == MP_INT:
        exponent = -<int> mp_decode_int(p)
    else:
        raise TypeError('unexpected exponent type: {}'.format(obj_type))

    length -= (&p[0][0] - first)
    first = &p[0][0]

    while first[0] == 0:
        first += 1  # skipping leading zeros

    sign = last[0] & 0xf  # extract sign
    if sign == 0x0a or sign == 0x0c or sign == 0x0e or sign == 0x0f:
        sign = 0
    else:
        sign = 1

    # decode digits
    digits_count = (last - first) * 2 + 1
    if first[0] & 0xf0 == 0:
        digits_count -= 1  # adjust for leading zero nibble

    digits = cpython.tuple.PyTuple_New(digits_count)

    if digits_count > 0:
        while True:
            nibble = (last[0] & 0xf0) >> 4  # left nibble first
            item = <object> <int> nibble
            cpython.Py_INCREF(item)
            cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)

            digits_count -= 1
            if digits_count == 0:
                break
            last -= 1

            nibble = last[0] & 0x0f  # right nibble
            item = <object> <int> nibble
            cpython.Py_INCREF(item)
            cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)

            digits_count -= 1
            if digits_count == 0:
                break

    p[0] += length

    return Decimal((<object> <int> sign, digits, <object> exponent))



from libc.stdint cimport uint32_t


cdef class MPInterval:
    cdef:
        public int year
        public int month
        public int week
        public int day
        public int hour
        public int min
        public int sec
        public int nsec
        public object adjust

cdef enum mp_interval_fields:
    MP_INTERVAL_FIELD_YEAR = 0
    MP_INTERVAL_FIELD_MONTH = 1
    MP_INTERVAL_FIELD_WEEK = 2
    MP_INTERVAL_FIELD_DAY = 3
    MP_INTERVAL_FIELD_HOUR = 4
    MP_INTERVAL_FIELD_MINUTE = 5
    MP_INTERVAL_FIELD_SECOND = 6
    MP_INTERVAL_FIELD_NANOSECOND = 7
    MP_INTERVAL_FIELD_ADJUST = 8

cdef uint32_t interval_len(MPInterval interval)
cdef char *interval_encode(char *p, MPInterval interval) except NULL
cdef MPInterval interval_decode(const char ** p, uint32_t length) except *



cimport cpython.datetime
from cpython.datetime cimport PyDateTimeAPI, datetime, datetime_tzinfo, timedelta_new
from libc.stdint cimport uint32_t
from libc.string cimport memcpy


cdef inline void datetime_zero(IProtoDateTime *dt):
    dt.seconds = 0
    dt.nsec = 0
    dt.tzoffset = 0
    dt.tzindex = 0

cdef inline uint32_t datetime_len(IProtoDateTime *dt):
    cdef uint32_t sz
    sz = sizeof(int64_t)
    if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
        return sz + DATETIME_TAIL_SZ
    return sz

cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL:
    store_u64(p, dt.seconds)
    p += sizeof(dt.seconds)
    if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
        memcpy(p, &dt.nsec, DATETIME_TAIL_SZ)
        p += DATETIME_TAIL_SZ
    return p

cdef int datetime_decode(
        const char ** p,
        uint32_t length,
        IProtoDateTime *dt
) except -1:
    delta = None
    tz = None

    dt.seconds = load_u64(p[0])
    p[0] += sizeof(dt.seconds)
    length -= sizeof(dt.seconds)

    if length == 0:
        return 0

    if length != DATETIME_TAIL_SZ:
        raise ValueError("invalid datetime size. got {} extra bytes".format(
            length
        ))

    dt.nsec = load_u32(p[0])
    p[0] += 4
    dt.tzoffset = load_u16(p[0])
    p[0] += 2
    dt.tzindex = load_u16(p[0])
    p[0] += 2

cdef void datetime_from_py(datetime ob, IProtoDateTime *dt):
    cdef:
        double ts
        int offset
    ts = <double> ob.timestamp()
    dt.seconds = <int64_t> ts
    dt.nsec = <int32_t> ((ts - <double> dt.seconds) * 1000000) * 1000
    if dt.nsec < 0:
        # correction for negative dates
        dt.seconds -= 1
        dt.nsec += 1000000000

    if datetime_tzinfo(ob) is not None:
        offset = ob.utcoffset().total_seconds()
        dt.tzoffset = <int16_t> (offset / 60)

cdef object datetime_to_py(IProtoDateTime *dt):
    cdef:
        double timestamp
        object tz

    tz = None

    if dt.tzoffset != 0:
        delta = timedelta_new(0, <int> dt.tzoffset * 60, 0)
        tz = timezone_new(delta)

    timestamp = dt.seconds + (<double> dt.nsec) / 1e9
    return PyDateTimeAPI.DateTime_FromTimestamp(
        <PyObject *>PyDateTimeAPI.DateTimeType,
        (timestamp,) if tz is None else (timestamp, tz),
        NULL,
    )



cdef class IProtoErrorStackFrame:
    cdef:
        readonly str error_type
        readonly str file
        readonly int line
        readonly str message
        readonly int err_no
        readonly int code
        readonly dict fields

cdef class IProtoError:
    cdef:
        readonly list trace

cdef IProtoError iproto_error_decode(const char ** b, bytes encoding)



import enum

from libc.stdint cimport int64_t, uint8_t, uint32_t, uint64_t


class Adjust(enum.IntEnum):
    """
    Interval adjustment mode for year and month arithmetic.
    """
    EXCESS = 0
    NONE = 1
    LAST = 2


cdef class MPInterval:
    def __cinit__(self,
                  int year=0,
                  int month=0,
                  int week=0,
                  int day=0,
                  int hour=0,
                  int min=0,
                  int sec=0,
                  int nsec=0,
                  object adjust=Adjust.NONE):
        self.year = year
        self.month = month
        self.week = week
        self.day = day
        self.hour = hour
        self.min = min
        self.sec = sec
        self.nsec = nsec
        self.adjust = adjust

    def __repr__(self):
        return (f"asynctnt.Interval("
                f"year={self.year}, "
                f"month={self.month}, "
                f"week={self.week}, "
                f"day={self.day}, "
                f"hour={self.hour}, "
                f"min={self.min}, "
                f"sec={self.sec}, "
                f"nsec={self.nsec}, "
                f"adjust={self.adjust!r}"
                f")")

    def __eq__(self, other):
        cdef:
            MPInterval other_interval

        if not isinstance(other, MPInterval):
            return False

        other_interval = <MPInterval> other

        return (self.year == other_interval.year
                and self.month == other_interval.month
                and self.week == other_interval.week
                and self.day == other_interval.day
                and self.hour == other_interval.hour
                and self.min == other_interval.min
                and self.sec == other_interval.sec
                and self.nsec == other_interval.nsec
                and self.adjust == other_interval.adjust
                )

cdef uint32_t interval_value_len(int64_t value):
    if value == 0:
        return 0

    if value > 0:
        return 1 + mp_sizeof_uint(<uint64_t> value)

    return 1 + mp_sizeof_int(value)

cdef char *interval_value_pack(char *data, mp_interval_fields field, int64_t value):
    if value == 0:
        return data

    data = mp_encode_uint(data, field)

    if value > 0:
        return mp_encode_uint(data, <uint64_t> value)

    return mp_encode_int(data, value)

cdef uint32_t interval_len(MPInterval interval):
    return (1
            + interval_value_len(interval.year)
            + interval_value_len(interval.month)
            + interval_value_len(interval.week)
            + interval_value_len(interval.day)
            + interval_value_len(interval.hour)
            + interval_value_len(interval.min)
            + interval_value_len(interval.sec)
            + interval_value_len(interval.nsec)
            + interval_value_len(<int64_t> interval.adjust.value)
            )

cdef char *interval_encode(char *data, MPInterval interval) except NULL:
    cdef:
        uint8_t fields_count

    fields_count = (<uint8_t>(interval.year != 0)
                    + <uint8_t>(interval.month != 0)
                    + <uint8_t>(interval.week != 0)
                    + <uint8_t>(interval.day != 0)
                    + <uint8_t>(interval.hour != 0)
                    + <uint8_t>(interval.min != 0)
                    + <uint8_t>(interval.sec != 0)
                    + <uint8_t>(interval.nsec != 0)
                    + <uint8_t>(interval.adjust != 0)
                    )
    data = mp_store_u8(data, fields_count)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_YEAR, <int64_t> interval.year)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_MONTH, <int64_t> interval.month)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_WEEK, <int64_t> interval.week)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_DAY, <int64_t> interval.day)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_HOUR, <int64_t> interval.hour)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_MINUTE, <int64_t> interval.min)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_SECOND, <int64_t> interval.sec)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_NANOSECOND, <int64_t> interval.nsec)
    data = interval_value_pack(data, MP_INTERVAL_FIELD_ADJUST, <int64_t> interval.adjust.value)
    return data

cdef MPInterval interval_decode(const char ** p,
                                   uint32_t length) except*:
    cdef:
        char *end
        MPInterval interval
        uint8_t fields_count
        int64_t value
        uint8_t field_type
        mp_type field_value_type

    end = p[0] + length
    fields_count = mp_load_u8(p)
    length -= sizeof(uint8_t)
    if fields_count > 0 and length < 2:
        raise ValueError("Invalid MPInterval length")

    interval = <MPInterval> MPInterval.__new__(MPInterval)

    # NONE is default but it will be encoded,
    # and because zeros are not encoded then we must set a zero value
    interval.adjust = Adjust.EXCESS

    for i in range(fields_count):
        field_type = mp_load_u8(p)
        value = 0
        field_value_type = mp_typeof(p[0][0])
        if field_value_type == MP_UINT:
            if mp_check_uint(p[0], end) > 0:
                raise ValueError(f"invalid uint. field_type: {field_type}")

        elif field_value_type == MP_INT:
            if mp_check_int(p[0], end) > 0:
                raise ValueError(f"invalid int. field_type: {field_type}")

        else:
            raise ValueError("Invalid MPInterval field value type")

        if mp_read_int64(p, &value) != 0:
            raise ValueError("Invalid MPInterval value")

        if field_type == MP_INTERVAL_FIELD_YEAR:
            interval.year = value
        elif field_type == MP_INTERVAL_FIELD_MONTH:
            interval.month = value
        elif field_type == MP_INTERVAL_FIELD_WEEK:
            interval.week = value
        elif field_type == MP_INTERVAL_FIELD_DAY:
            interval.day = value
        elif field_type == MP_INTERVAL_FIELD_HOUR:
            interval.hour = value
        elif field_type == MP_INTERVAL_FIELD_MINUTE:
            interval.min = value
        elif field_type == MP_INTERVAL_FIELD_SECOND:
            interval.sec = value
        elif field_type == MP_INTERVAL_FIELD_NANOSECOND:
            interval.nsec = value
        elif field_type == MP_INTERVAL_FIELD_ADJUST:
            interval.adjust = Adjust(<int> value)
        else:
            raise ValueError(f"Invalid MPInterval field type {field_type}")

    return interval
